require "defines"
require "migration"

local SLIPSTREAM_VERSION = "1.1.0"
local sschests = {}
local energy_per_action = 5000

local next = next
local north = defines.direction.north
local east = defines.direction.east
local south = defines.direction.south
local west = defines.direction.west

game.oninit(function()
	onLoad()
end)

game.onload(function()
	onLoad()
end)

game.onsave(function()
	glob.slipstream.sschests = sschests
end)

game.onevent(defines.events.onbuiltentity, function(event)
	if string.find(event.createdentity.name, "^slipstream%-chest%-[%a%-]+$") then
		slipstreamBuilt(event.createdentity)
	end
end)

game.onevent(defines.events.onrobotbuiltentity, function(event)
	if string.find(event.createdentity.name, "^slipstream%-chest%-[%a%-]+$") then
		slipstreamBuilt(event.createdentity)
	end
end)

game.onevent(defines.events.ontick, function(event)
	if string.match(game.tick,'^%d+(0)$') then
		for i,v in ipairs(sschests) do
			if v.entity.valid then
				chestCollection(v)
				if v.output_config.limit > 0 then chestSpawn(v) end
			else
				entityRemoved()
			end
		end
	end
end)

function onLoad()
	if not glob.slipstream then
		glob.slipstream = {}
		glob.slipstream.sschests = {}
		glob.slipstream.version = SLIPSTREAM_VERSION
	end
	
	if glob.slipstream.version ~= SLIPSTREAM_VERSION then
		slipstreamMigration(glob.slipstream.version, SLIPSTREAM_VERSION)
	end
	
	sschests = glob.slipstream.sschests
end

function entityRemoved()
	for i,v in ipairs(sschests) do
		if not v.entity.valid then
			v.battery.destroy()
			table.remove(sschests,i)
		end
	end
end

-- Construct the data table for each chest as it's built.
function slipstreamBuilt(entity)
	local chest_name = entity.name
	
	local new_chest = {}
	new_chest.output_config = {}
	
	-- Record position of chest and search area.
	local pos = {x = entity.position.x, y = entity.position.y}
	local search = {{pos.x - 0.5,pos.y - 0.5},{pos.x + 0.5, pos.y + 0.5}}
	
	entity.destroy() -- Remove initial chest.
	
	-- Creates the accumulator entity that serves as the chest's power supply.
	game.createentity{name="slipstream-chest-battery", position = pos, force = game.forces.player}
	
	-- Recreate the chest on top of the accumulator.
	game.createentity{name = chest_name, position = pos, force = game.forces.player}
	
	-- Find the entity references for the battery/chest.
	local t = game.findentitiesfiltered{area = search, name = "slipstream-chest-battery"}
	new_chest.battery = t[1]
	new_chest.battery.minable = false
	new_chest.battery.destructible = false
	local t = game.findentitiesfiltered{area = search, name = chest_name}
	new_chest.entity = t[1]
	new_chest.inventory = t[1].getinventory(defines.inventory.chest)
	
	-- Add coordinates for collection/spawning of items.
	new_chest.coords_collection = configureChestCollection(pos)
	
	-- Autoconfigure output based on dummy's direction and presence of transport belt.
	new_chest.coords_spawn, new_chest.coords_spawn_catchment, new_chest.output_config.limit = configureChestSpawn(pos)
	new_chest.output_config.target = 1
	
	table.insert(sschests, new_chest)
	
	--[[
	debugLog("Chest: " .. tostring(new_chest.entity.name) .. " " .. table.tostring(pos))
	debugLog("Battery: " .. tostring(new_chest.battery.name))
	debugLog("Inventory: " .. tostring(new_chest.inventory))
	if new_chest.coords_spawn[1] ~= nil then
		debugLog("Spawn Pos 1: " .. new_chest.coords_spawn[1].x .. ", " .. new_chest.coords_spawn[1].y)
		debugLog("Spawn Pos 2: " .. new_chest.coords_spawn[2].x .. ", " .. new_chest.coords_spawn[2].y)
	end
	debugLog(new_chest.output_config.limit)--]]
end

-- Configure the area that the chest will search for new items.
function configureChestCollection(pos)
	return {
		{pos.x - 0.5,pos.y - 0.5},
		{pos.x + 0.5,pos.y + 0.5}
	}
end

-- Configure the eight locations the chest will attempt to place items down on (in all cardinal directions) if configured to do so.
function configureChestSpawn(pos)
	local spawn_coords = { -- Set up output coordinates.
		[0] = {x = pos.x - 0.25, y = pos.y - 0.64},
		[1] = {x = pos.x + 0.25, y = pos.y - 0.64},
		[2] = {x = pos.x + 0.64, y = pos.y - 0.25},
		[3] = {x = pos.x + 0.64, y = pos.y + 0.25},
		[4] = {x = pos.x + 0.25, y = pos.y + 0.64},
		[5] = {x = pos.x - 0.25, y = pos.y + 0.64},
		[6] = {x = pos.x - 0.75, y = pos.y + 0.25},
		[7] = {x = pos.x - 0.75, y = pos.y - 0.25}
	}
	
	local beltscan_coords = { -- Points to search for transport belts.
		[north] = {{pos.x - 0.1, pos.y - 1.1},{pos.x + 0.1, pos.y - 0.9}},
		[east] = {{pos.x + 0.9, pos.y - 0.1},{pos.x + 1.1, pos.y + 0.1}},
		[south] = {{pos.x - 0.1, pos.y + 0.9},{pos.x + 0.1, pos.y + 1.1}},
		[west] = {{pos.x - 1.1, pos.y - 0.1},{pos.x - 0.9, pos.y + 0.1}}
	}
	
	local directions = {north, east, south, west} -- For the for loop.
	local facing_directions = {[north] = south, [east] = west, [south] = north, [west] = east} -- If a found belt is facing in this direction, ignore it.
	local spawns_list = {}
	local limit = 0
	
	for i,v in ipairs(directions) do
		local belt = game.findentitiesfiltered{area = beltscan_coords[v], type = "transport-belt"} -- Search for a transport belt.
		if belt[1] ~= nil and belt[1].direction ~= facing_directions[v] then -- If belt is found, add two spawn locations to coords_spawn and +2 to output limit counter.
			table.insert(spawns_list,spawn_coords[v])
			table.insert(spawns_list,spawn_coords[v+1])
			limit = limit + 2
		end
	end
	
	-- Creates a pair of small search squares around where output items are deposited. These are searched to determine if anything is in them, since canplaceentity seems bugged.
	local spawns_catchment = {}
	for i,v in ipairs(spawns_list) do
		spawns_catchment[i] = {{v.x - 0.2, v.y - 0.2},{v.x + 0.2, v.y + 0.2}}
	end
	
	-- TODO: Add check and configuration for belt_to_ground.
	
	return spawns_list, spawns_catchment, limit
end

function chestCollection(chest) -- Pick up stuff fed towards the square the chest is placed in.
	local pickup = game.findentitiesfiltered{area = chest.coords_collection, name = "item-on-ground"}
	
	if next(pickup) ~= nil then
		local imax = #pickup
		local i = 1
		
		while i <= imax and chest.battery.energy >= energy_per_action do
			if chest.inventory.caninsert(pickup[i].stack) then
				chest.inventory.insert(pickup[i].stack)
				pickup[i].destroy()
				chest.battery.energy = chest.battery.energy - energy_per_action
				i = i + 1
			else
				return
			end
		end
	end
end

-- Place items in inventory on spawn positions.
function chestSpawn(chest)
	local output_count = 0 -- ensures output is limited to one for each valid output.
	local target = chest.output_config.target -- copied into local var to reduce typing.
	
	while output_count < chest.output_config.limit and not chest.inventory.isempty() and chest.battery.energy >= energy_per_action do
		if target > chest.output_config.limit then
			target = 1
		end
		
		local pos = chest.coords_spawn[target] -- coordinates to spawn item at.
		local catchment = chest.coords_spawn_catchment[target]
		local inventory = chest.inventory.getcontents()
		local itemname = next(inventory)
		
		--[[if itemname == "any" then
			for i in ipairs(chest.output_config.filter) do
				if inventory[i] ~= nil then
					table.remove(inventory, i)
				end
			end
			
			itemname = next(inventory)
		end--]]
		
		-- game.canplaceentity{name = "item-on-ground", position = pos, stack = {name = itemname, count = 1}}
		-- next(game.findentitiesfiltered{area = catchment, name="item-on-ground"}) == nil
		if next(game.findentitiesfiltered{area = catchment, name="item-on-ground"}) == nil then
			game.createentity{name = "item-on-ground", position = pos, stack = {name = itemname, count = 1}, force = game.forces.player}
			chest.inventory.remove{name = itemname, count = 1}
			chest.battery.energy = chest.battery.energy - energy_per_action
		end
		
		target = target + 1
		output_count = output_count + 1
	end
	
	chest.output_config.target = target
	
	--[[for i,v in pairs(storage.spawn) do
		if not storage.inventory.isempty() then
			local itemname = "foo"
			do	-- Compiles a list of names of items sitting in the chest.
				local keyset = {}
				local n=0
				
				for k,_ in pairs(storage.inventory.getcontents()) do
					n=n+1
					keyset[n]=k
				end
				
				itemname = keyset[1]
			end
			
			spawnposentity = game.findentitiesfiltered{area = {storage.spawncatchment[i][1], storage.spawncatchment[i][2]}, name="item-on-ground"}
			
			if next(spawnposentity) == nil then -- Tests whether anything is in the area where items are deposited by the chest, if not, proceed to putting stuff down.
				game.createentity{name="item-on-ground", position = v, stack = {name = itemname, count = 1}}
				storage.inventory.remove{name = itemname, count = 1}
			end
		end
	end--]]
end

--[[ These aren't relevant to the mod except as debugging tools.
function debugLog(message)
	if true then -- set for debug
		game.player.print(message)
	end
end

function table.val_to_str ( v )
  if "string" == type( v ) then
    v = string.gsub( v, "\n", "\\n" )
    if string.match( string.gsub(v,"[^'\"]",""), '^"+$' ) then
      return "'" .. v .. "'"
    end
    return '"' .. string.gsub(v,'"', '\\"' ) .. '"'
  else
    return "table" == type( v ) and table.tostring( v ) or
      tostring( v )
  end
end

function table.key_to_str ( k )
  if "string" == type( k ) and string.match( k, "^[_%a][_%a%d]*$" ) then
    return k
  else
    return "[" .. table.val_to_str( k ) .. "]"
  end
end

function table.tostring( tbl )
  local result, done = {}, {}
  for k, v in ipairs( tbl ) do
    table.insert( result, table.val_to_str( v ) )
    done[ k ] = true
  end
  for k, v in pairs( tbl ) do
    if not done[ k ] then
      table.insert( result,
        table.key_to_str( k ) .. "=" .. table.val_to_str( v ) )
    end
  end
  return "{" .. table.concat( result, "," ) .. "}"
end --]]
